import re

from nuclear.sublog import log

from .request import HttpRequest


def replace_request_path(request: HttpRequest, match_regex: str, replace_regex: str) -> HttpRequest:
    """
    Replace request path applying Regex.
    Example:
        replace_request_path(request, r'^/path/(.+?)(/[a-z]+)(/.*)', r'\3')
    """
    matcher = re.compile(match_regex)
    if matcher.match(request.path):
        request.path = matcher.sub(replace_regex, request.path)
        log.debug('request path transformed', path=request.path)
    return request



from nuclear import CliBuilder, parameter, argument, flag
from nuclear.types.boolean import boolean

from .setup import setup_proxy
from .version import __version__


def main():
    CliBuilder('x-man', run=setup_proxy, help_on_empty=True, version=__version__,
               help='HTTP proxy recording & replaying requests').has(
        argument('dst_url', help='destination base url', required=False, default='http://127.0.0.1:8000'),
        parameter('listen_port', help='listen port for incoming requests', type=int, default=8080),
        parameter('listen_ssl', help='enable https on listening side', type=boolean, default=True),
        parameter('record', help='enable recording requests & responses', type=boolean, default=False),
        parameter('record_file', help='filename with recorded requests', default='tape.json'),
        parameter('replay', help='return cached results if found', type=boolean, default=False),
        parameter('replay_throttle', type=boolean, default=False,
                  help='throttle response if too many requests are made'),
        parameter('replay_clear_cache', help='enable clearing cache periodically', type=boolean, default=False),
        parameter('replay_clear_cache_seconds', help='clearing cache interval in seconds', type=int, default=1 * 60),
        parameter('allow_chunking', help='enable sending response in chunks', type=boolean, default=True),
        parameter('ext', help='load extensions from Python file'),
        flag('verbose', 'v', help='show more details in output'),
    ).run()



import requests
import urllib3
from nuclear.sublog import log, wrap_context

from .request import HttpRequest
from .response import HttpResponse

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


def proxy_request(request: HttpRequest, base_url: str) -> HttpResponse:
    with wrap_context('proxying to destination', base_url=base_url):
        url = f'{base_url}{request.path}'
        log.debug(f'>> proxying to', url=url)
        response = requests.request(request.method, url, verify=False, allow_redirects=False, stream=False,
                                    timeout=10, headers=request.headers, data=request.content)
        content: bytes = response.content
        return HttpResponse(status_code=response.status_code, headers=dict(response.headers), content=content)



import json
import os
import zlib
from datetime import datetime
from pathlib import Path
from typing import Dict, Tuple, List, Any

from dataclasses import dataclass, is_dataclass, asdict
from nuclear.sublog import log

from .config import Config
from .extension import Extensions
from .request import HttpRequest
from .response import HttpResponse


@dataclass
class CacheEntry(object):
    request: HttpRequest
    response: HttpResponse

    @staticmethod
    def from_json(data: dict) -> 'CacheEntry':
        return CacheEntry(
            request=HttpRequest.from_json(data.get('request')),
            response=HttpResponse.from_json(data.get('response')),
        )


class RequestCache(object):
    def __init__(self, extensions: Extensions, config: Config):
        self.extensions: Extensions = extensions
        self.config: Config = config
        self.cache: Dict[int, CacheEntry] = self._init_request_cache()

    def _init_request_cache(self) -> Dict[int, CacheEntry]:
        if self.config.record_file and os.path.isfile(self.config.record_file):
            txt = Path(self.config.record_file).read_text()
            if not txt:
                return {}
            entries = json.loads(txt)
            loaded_cache = {}
            for entry in entries:
                parsed_entry = CacheEntry.from_json(entry)
                request_hash = self._request_hash(parsed_entry.request)
                loaded_cache[request_hash] = parsed_entry
            log.debug(f'CACHE: loaded request-response pairs', record_file=self.config.record_file,
                      read_entries=len(entries), distinct_entries=len(loaded_cache))
            return loaded_cache
        return {}

    def has_cached_response(self, request: HttpRequest) -> bool:
        return self.config.replay and self._can_be_cached(request) and self._request_hash(request) in self.cache

    def _can_be_cached(self, request: HttpRequest) -> bool:
        if self.extensions.can_be_cached is None:
            return True
        return self.extensions.can_be_cached(request)

    def get(self, request_hash: int) -> CacheEntry:
        return self.cache[request_hash]

    def replay_response(self, request: HttpRequest) -> HttpResponse:
        request_hash = self._request_hash(request)
        if self.config.replay_throttle:
            log.debug('CACHE: Throttled response', hash=request_hash)
            return too_many_requests_response
        log.debug('CACHE: Found cached response', hash=request_hash)
        return self.cache[request_hash].response

    def clear_old(self):
        if not self.config.replay_clear_cache:
            return
        to_remove = []
        now_timestamp: float = now_seconds()
        for request_hash, entry in self.cache.items():
            if now_timestamp - entry.request.timestamp > self.config.replay_clear_cache_seconds:
                to_remove.append(request_hash)
        for request_hash in to_remove:
            del self.cache[request_hash]
        if to_remove:
            log.debug('CACHE: cleared old cache entries', removed=len(to_remove))

    def saving_enabled(self, request: HttpRequest) -> bool:
        return (self.config.record or self.config.replay) and self._can_be_cached(request)

    def save_response(self, request: HttpRequest, response: HttpResponse):
        request_hash = self._request_hash(request)
        if request_hash not in self.cache:
            self.cache[request_hash] = CacheEntry(request, response)
            if self.config.record and self.config.record_file:
                serializable = list(self.cache.values())
                txt = json.dumps(serializable, sort_keys=True, indent=4, cls=EnhancedJSONEncoder)
                Path(self.config.record_file).write_text(txt)
            log.debug(f'+ CACHE: new request-response recorded', hash=request_hash, total_entries=len(self.cache))

    def _request_hash(self, request: HttpRequest) -> int:
        traits_str = str(self._request_traits(request))
        return zlib.adler32(traits_str.encode())

    def _request_traits(self, request: HttpRequest) -> Tuple:
        if self.extensions.cache_request_traits is None:
            return default_request_traits(request)
        return self.extensions.cache_request_traits(request)


def default_request_traits(request: HttpRequest) -> Tuple:
    return request.method, request.path, request.content, sorted_dict_trait(request.headers)


def sorted_dict_trait(d: Dict[str, Any]) -> List[Tuple[str, Any]]:
    return sorted([(k, v) for k, v in d.items()], key=lambda t: t[0])


too_many_requests_response = HttpResponse(status_code=429, headers={}, content=b'')


def now_seconds() -> float:
    return datetime.now().timestamp()


class EnhancedJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if is_dataclass(obj):
            return asdict(obj)
        if isinstance(obj, bytes):
            return obj.decode('utf-8')
        return super().default(obj)



from importlib.machinery import SourceFileLoader
from typing import Callable, Tuple, Optional

from dataclasses import dataclass, fields
from nuclear.sublog import log

from .config import Config
from .request import HttpRequest
from .response import HttpResponse


@dataclass
class Extensions(object):
    transform_request: Optional[Callable[[HttpRequest], HttpRequest]] = None
    transform_response: Optional[Callable[[HttpRequest, HttpResponse], HttpResponse]] = None
    can_be_cached: Optional[Callable[[HttpRequest], bool]] = None
    cache_request_traits: Optional[Callable[[HttpRequest], Tuple]] = None
    override_config: Optional[Callable[[Config], None]] = None


def load_extensions(extension_path: str) -> Extensions:
    if not extension_path:
        return Extensions()

    log.debug(f'loading extensions', path=extension_path)
    ext = Extensions()
    ext_module = SourceFileLoader("x_man.transformer", extension_path).load_module()
    ext_names = [field.name for field in fields(Extensions)]
    for ext_name in ext_names:
        if hasattr(ext_module, ext_name):
            ext_value = getattr(ext_module, ext_name)
            if ext_value:
                setattr(ext, ext_name, ext_value)
                log.debug(f'loaded {ext_name} extension')

    return ext



import json
from typing import Dict, Callable, Any, Optional

from dataclasses import dataclass
from nuclear.sublog import log


@dataclass
class HttpRequest(object):
    requestline: str
    method: str
    path: str
    headers: Dict[str, str]
    content: bytes
    client_addr: str
    client_port: int
    timestamp: float

    @staticmethod
    def from_json(data: dict) -> 'HttpRequest':
        data['content'] = data.get('content').encode('utf-8')
        return HttpRequest(**data)

    def log(self, verbose: bool):
        if not verbose:
            log.info(f'< Incoming: {self.requestline}')
        elif self.content:
            log.info(f'< Incoming: {self.requestline}', headers=self.headers, content='\n'+self.content.decode('utf-8'))
        else:
            log.info(f'< Incoming: {self.requestline}', headers=self.headers)

    def transform(self, transformer: Optional[Callable[['HttpRequest'], 'HttpRequest']]) -> 'HttpRequest':
        if transformer is None:
            return self
        cloned = HttpRequest(
            requestline=self.requestline,
            method=self.method,
            path=self.path,
            headers=self.headers,
            content=self.content,
            client_addr=self.client_addr,
            client_port=self.client_port,
            timestamp=self.timestamp,
        )
        return transformer(cloned)

    def json(self) -> Any:
        if len(self.content) == 0:
            return None
        return json.loads(self.content)



import json
from http.client import responses
from typing import Dict, Callable, Any, Optional

from dataclasses import dataclass
from nuclear.sublog import log

from .request import HttpRequest


@dataclass
class HttpResponse(object):
    status_code: int
    headers: Dict[str, str]
    content: bytes

    def log(self, prefix: str, verbose: bool) -> 'HttpResponse':
        status = f'{self.status_code} {responses[self.status_code]}'
        if verbose:
            log.debug(f'{prefix}', status=status, headers=self.headers, content='\n' + self.content.decode())
        else:
            log.debug(f'{prefix}', status=status)
        return self

    @staticmethod
    def from_json(data: dict) -> 'HttpResponse':
        data['content'] = data.get('content').encode()
        return HttpResponse(**data)

    def transform(self, transformer: Optional[Callable[[HttpRequest, 'HttpResponse'], 'HttpResponse']],
                  request: HttpRequest) -> 'HttpResponse':
        if transformer is None:
            return self
        cloned = HttpResponse(
            status_code=self.status_code,
            headers=self.headers,
            content=self.content,
        )
        return transformer(request, cloned)

    def set_content(self, content: str) -> 'HttpResponse':
        self.content = content.encode()
        self.headers['Content-Length'] = str(len(self.content))
        return self

    def set_json(self, obj: Any) -> 'HttpResponse':
        self.content = json.dumps(obj).encode()
        self.headers['Content-Length'] = str(len(self.content))
        self.headers['Content-Type'] = 'application/json'
        return self

    def json(self) -> Any:
        if len(self.content) == 0:
            return None
        return json.loads(self.content)



import json
import sys
from http.server import SimpleHTTPRequestHandler
from typing import Dict, Iterable, Sequence

from nuclear.sublog import log, log_error, wrap_context

from .cache import RequestCache, now_seconds
from .config import Config
from .extension import Extensions
from .proxy import proxy_request
from .request import HttpRequest
from .response import HttpResponse


class RequestHandler(SimpleHTTPRequestHandler):
    extensions: Extensions
    config: Config
    cache: RequestCache

    def handle_request(self):
        with log_error():
            with wrap_context('handling request'):
                self.connection.settimeout(10)
                incoming_request = self.incoming_request()
                incoming_request.log(self.config.verbose)
                response_0 = self.generate_response(incoming_request)
                response = response_0.transform(self.extensions.transform_response, incoming_request)
                if response != response_0:
                    response.log('response transformed', self.config.verbose)
                self.respond_to_client(response)

    def incoming_request(self) -> HttpRequest:
        with wrap_context('building incoming request'):
            headers_dict = {parsed_line[0]: parsed_line[1] for parsed_line in self.headers.items()}
            method = self.command.lower()
            content_len = int(headers_dict.get('Content-Length', 0))
            content: bytes = self.rfile.read(content_len) if content_len else b''
            return HttpRequest(requestline=self.requestline, method=method, path=self.path,
                               headers=headers_dict, content=content,
                               client_addr=self.client_address[0], client_port=self.client_address[1],
                               timestamp=now_seconds())

    def generate_response(self, request_0: HttpRequest) -> HttpResponse:
        with wrap_context('generating response'):
            request = request_0.transform(self.extensions.transform_request)
            if request != request_0:
                log.debug('request transformed')

            self.cache.clear_old()
            if self.cache.has_cached_response(request):
                return self.cache.replay_response(request).log('> returning', self.config.verbose)

            response: HttpResponse = proxy_request(request, base_url=f'{self.config.dst_url}')
            response.log('<< received', self.config.verbose)

            if self.cache.saving_enabled(request):
                self.cache.save_response(request, response)

            return response

    def respond_to_client(self, response: HttpResponse):
        with wrap_context('responding back to client'):
            self.send_response_only(response.status_code)

            if 'Content-Encoding' in response.headers:
                del response.headers['Content-Encoding']
                log.debug('removing Content-Encoding header')

            if 'Content-Length' not in response.headers and response.content:
                response.headers['Content-Length'] = str(len(response.content))
                log.debug('adding missing Content-Length header')

            for name, value in response.headers.items():
                self.send_header(name, value)
            self.end_headers()

            if self.config.allow_chunking and response.headers.get('Transfer-Encoding') == 'chunked':
                self.send_chunked_response(chunks(response.content, 512))
            else:
                self.wfile.write(response.content)
            self.close_connection = True
            log.debug('> response sent', client_addr=self.client_address[0], client_port=self.client_address[1])

    def send_chunked_response(self, content_chunks: Iterable[bytes]):
        for chunk in content_chunks:
            tosend = ('%X' % len(chunk)).encode('utf-8') + b'\r\n' + chunk + b'\r\n'
            self.wfile.write(tosend)
        self.wfile.write('0\r\n\r\n'.encode('utf-8'))

    def respond_json(self, response: Dict):
        self.send_response(200)
        self.send_header('Content-Type', 'application/json')
        self.end_headers()
        self.wfile.write(json.dumps(response).encode())

    def do_GET(self):
        self.dev_var() or self.handle_request()

    def do_POST(self):
        self.dev_var() or self.handle_request()

    def do_PUT(self):
        self.handle_request()

    def do_DELETE(self):
        self.handle_request()

    def do_HEAD(self):
        self.handle_request()

    def dev_var(self) -> bool:
        if self.path != '/dev/var':
            return False
        content_len = int(self.headers.get('Content-Length', 0))
        content = self.rfile.read(content_len) if content_len else ''
        if not content:
            return False
        payload = json.loads(content)
        var_name = payload.get('name')
        if not var_name:
            return False
        if self.command.lower() == 'POST':
            var_value = payload.get('value')
            setattr(sys.modules[__name__], var_name, var_value)
            log.debug(f'dev variable set', name=var_name, value=var_value)
        var_value = getattr(sys.modules[__name__], var_name)
        self.respond_json({'name': var_name, 'value': var_value})
        return True


def chunks(lst: Sequence, n: int) -> Iterable:
    for i in range(0, len(lst), n):
        yield lst[i:i + n]



import setuptools
from x_man import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fh:
    install_requires = fh.read().splitlines()

setuptools.setup(
    name="x-man",
    version=__version__,
    author="igrek51",
    author_email="igrek51.dev@gmail.com",
    description="HTTP proxy recording & replaying requests",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/igrek51/x-man",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=install_requires,
    license='MIT',
    python_requires='>=3.6.0',
    entry_points={
        "console_scripts": [
            "x-man = x_man:main",
        ],
    },
)



from dataclasses import dataclass


@dataclass
class Config(object):
    listen_port: int
    listen_ssl: bool
    dst_url: str
    record: bool
    record_file: str
    replay: bool
    replay_throttle: bool
    replay_clear_cache: bool
    replay_clear_cache_seconds: int
    allow_chunking: bool
    verbose: bool



import ssl
from socketserver import TCPServer

from nuclear.sublog import logerr, wrap_context, log

from .cache import RequestCache
from .config import Config
from .extension import load_extensions
from .handler import RequestHandler


def setup_proxy(listen_port: int, listen_ssl: bool, dst_url: str, record: bool, record_file: str, replay: bool,
                replay_throttle: bool, replay_clear_cache: bool, replay_clear_cache_seconds: int, allow_chunking: bool,
                ext: str, verbose: bool):
    with logerr():
        with wrap_context('initialization'):
            extensions = load_extensions(ext)
            config = Config(
                listen_port=listen_port,
                listen_ssl=listen_ssl,
                dst_url=dst_url,
                record=record,
                record_file=record_file,
                replay=replay,
                replay_throttle=replay_throttle,
                replay_clear_cache=replay_clear_cache,
                replay_clear_cache_seconds=replay_clear_cache_seconds,
                allow_chunking=allow_chunking,
                verbose=verbose,
            )
            if extensions.override_config:
                extensions.override_config(config)

            RequestHandler.extensions = extensions
            RequestHandler.config = config
            RequestHandler.cache = RequestCache(extensions, config)

            TCPServer.allow_reuse_address = True
            httpd = TCPServer(("", config.listen_port), RequestHandler)
            if config.listen_ssl:
                httpd.socket = ssl.wrap_socket(httpd.socket, certfile='./dev-cert.pem', server_side=True)
            scheme = 'HTTPS' if config.listen_ssl else 'HTTP'
            log.info(f'Listening on {scheme} port {config.listen_port}...')
            try:
                httpd.serve_forever()
            finally:
                httpd.server_close()
